;;; -*- Mode: Lisp -*-
;;; Ystok-Local-Time - Calls to C run-time library libc on *nix and Windows
;;; Copyright (c) 2002-2014 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :ystok.local-time)

(defparameter *libc-module* #+win32 "msvcrt"
                            #-win32 "libc")

(eval-when (:compile-toplevel :load-toplevel :execute)

(ffc:define-foreign-type time-t :signed-long)

(ffc:define-foreign-struct libc-tm
  (tm-sec   :signed-long)			; seconds
  (tm-min   :signed-long)			; minutes
  (tm-hour  :signed-long)			; hours
  (tm-mday  :signed-long)			; day of month
  (tm-mon   :signed-long)			; month
  (tm-year  :signed-long)			; year
  (tm-wday  :signed-long)			; day of week
  (tm-yday  :signed-long)			; day of year
  (tm-isdst :signed-long))			; daylight savings time
) ; eval-when

(ffc:define-foreign-function (libc-mktime "mktime")
  ((tm (:pointer libc-tm)))
  time-t
  :calling-convention :cdecl :module *libc-module*)

(ffc:define-foreign-function (libc-gmtime "gmtime")
  ((time-ptr (:pointer time-t)))				; (:reference-pass time-t)
  (:pointer libc-tm)
  :calling-convention :cdecl :module *libc-module*)

;; Misfunctions on Windows NT? (always return the same as gmtime: tm-isdst = 0)
(ffc:define-foreign-function (libc-localtime "localtime")
  ((time-ptr (:pointer time-t)))				; (:reference-pass time-t)
  (:pointer libc-tm)
  :calling-convention :cdecl :module *libc-module*)

(defun unix-mktime (year month day &optional (hour 0) (min 0) (sec 0))
 ;; Interpret arguments as a local time parts.
 ;; Value: Unix time.
 ;; Set the tm_isdst field to 0 to indicate that standard time is in effect, 
 ;; or to a value greater than 0 to indicate that daylight savings time is in effect,
 ;; or to a value less than zero to have the C run-time library code compute 
 ;; whether standard time or daylight savings time is in effect. 
 ;; Surprise: Works correctly! Assumes that input parameters correspond to Windows
 ;;           regional settings of timezone and DST (USA DST rules?)
 ;;
  (declare (type integer year month day hour min sec))
  (ffc:with-foreign ((tm libc-tm))
    (setf (ffc:%slot tm libc-tm tm-year) year
          (ffc:%slot tm libc-tm tm-mon)  month
          (ffc:%slot tm libc-tm tm-mday) day
          (ffc:%slot tm libc-tm tm-hour) hour
          (ffc:%slot tm libc-tm tm-min)  min
          (ffc:%slot tm libc-tm tm-sec)  sec
          (ffc:%slot tm libc-tm tm-yday) 0
          (ffc:%slot tm libc-tm tm-wday) 0
          (ffc:%slot tm libc-tm tm-isdst) -1)		; let compute itself
    (libc-mktime tm)))

(defun unix-gmtime (unix-time)
  (declare (type integer unix-time))
  (ffc:with-foreign ((time-ptr time-t :initial-element unix-time))
    (let ((tm (libc-gmtime time-ptr))) 
      (values (ffc:%slot tm libc-tm tm-sec)
              (ffc:%slot tm libc-tm tm-min)
              (ffc:%slot tm libc-tm tm-hour)
              (ffc:%slot tm libc-tm tm-mday)
              (ffc:%slot tm libc-tm tm-mon)
              (ffc:%slot tm libc-tm tm-year)
              (ffc:%slot tm libc-tm tm-wday)		; day of week
              (ffc:%slot tm libc-tm tm-yday)		; day of year
              (ffc:%slot tm libc-tm tm-isdst)))))		; daylight savings time = 0

(defun unix-localtime (unix-time)
  (declare (type integer unix-time))
  (ffc:with-foreign ((time-ptr time-t :initial-element unix-time))
    (let ((tm (libc-localtime time-ptr)))
      (values (ffc:%slot tm libc-tm tm-sec)
              (ffc:%slot tm libc-tm tm-min)
              (ffc:%slot tm libc-tm tm-hour)
              (ffc:%slot tm libc-tm tm-mday)
              (ffc:%slot tm libc-tm tm-mon)
              (ffc:%slot tm libc-tm tm-year)
              (ffc:%slot tm libc-tm tm-wday)		; day of week
              (ffc:%slot tm libc-tm tm-yday)		; day of year
              (ffc:%slot tm libc-tm tm-isdst)))))		; daylight savings time = 0

;;; Timezone initialization - sets time environment variables of C RTL
(ffc:define-foreign-function (libc-tzset "_tzset") () :void
  :calling-convention :cdecl :module *libc-module*)

(eval-when (:load-toplevel :execute)
 (libc-tzset))						; to initialize library

;; Current timezone values:
;; - difference in seconds between coordinated universal time and local time
;; - daylight saving
(ffc:define-foreign-variable (libc-timezone "_timezone") time-t)
(ffc:define-foreign-variable (libc-daylight "_daylight") :signed-long)

(defun unix-timezone (&optional unix-time)
 ;;; Returns information about this timezone relative to moment of UNIX-TIME.
  ;; Arg: UNIX-TIME If null, assume the current time.
  ;; Values: 1) This zone displacement as seconds East of UCT.
  ;;         2) Boolean daylight-saving-p.
  (if unix-time
      ;; Beware of reusing the same C structure when returning results for both calls
      (ffc:with-foreign ((time-ptr time-t :initial-element unix-time))
        (let* ((localtime (libc-localtime time-ptr))
               (tm-sec    (ffc:%slot localtime libc-tm tm-sec))
               (tm-min    (ffc:%slot localtime libc-tm tm-min))
               (tm-year   (ffc:%slot localtime libc-tm tm-year))
               (tm-hour   (ffc:%slot localtime libc-tm tm-hour))
               (tm-yday   (ffc:%slot localtime libc-tm tm-yday))
               (tm-isdst  (ffc:%slot localtime libc-tm tm-isdst))
               (gmtime (libc-gmtime time-ptr))
               (year   (- tm-year (ffc:%slot gmtime libc-tm tm-year))))
          (declare (type fixnum tm-year tm-hour tm-yday tm-isdst year))
          (values (+ (- tm-sec (ffc:%slot gmtime libc-tm tm-sec))
                     (* 60
                        (+ (- tm-min (ffc:%slot gmtime libc-tm tm-min))
                           (* 60
                              (+ (- tm-hour (ffc:%slot gmtime libc-tm tm-hour))
                                 (* 24
                                    (if (zerop year)		; tm-year's are the same
                                        (- tm-yday (ffc:%slot gmtime libc-tm tm-yday))
                                        (signum year))))))))
                  (plusp tm-isdst))))
                ;(case tm-isdst (1 t) (0 nil) (otherwise :unknown))))
      ;; else calculate offset for the current time
      (let ((tm-isdst (/= libc-daylight 0)))
        (values (- (if tm-isdst 3600 0) libc-timezone)
                tm-isdst))))


#||;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;; EVALUATION ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

#+unused
(ffc:define-foreign-variable (libc-tzname "_tzname") (:c-array (:pointer :char) 2))

(ffc:%cstring-lisp (ffc:%svref libc-tzname 0))	;=> "Russian Standard Time"
(fli:convert-from-foreign-string (fli:foreign-aref libc-tzname 0))

(ffc:%cstring-lisp (ffc:%svref libc-tzname 1))	;=> "Russian Daylight Time"
(fli:convert-from-foreign-string (fli:foreign-aref libc-tzname 1))

(fli:define-foreign-function (msvc-time "time")
    ((ptime (:reference-pass time-t :allow-null t)))
  :module (load-time-value *libc-module*)
  :result-type time-t
  :calling-convention :cdecl)

;;; OLD STUFF
;;; $Id: unix.lisp,v 1.10 2002/12/04 17:11:26 jesse Exp $
;;; Copyright (c) 2000, 2001 onShore Development, Inc.

#+cmu

(def-alien-type time-t long)

(def-alien-type nil
    (struct time
            (tm-sec   int)              ; seconds
            (tm-min   int)              ; minutes
            (tm-hour  int)              ; hours
            (tm-mday  int)              ; day of month
            (tm-mon   int)              ; month
            (tm-year  int)              ; year
            (tm-wday  int)              ; day of week
            (tm-yday  int)              ; day of year
            (tm-isdst int)))            ; daylight savings time

(defun libc-gmtime (time)
  (let ((fn (extern-alien "gmtime" (function (* (struct time)) (* time-t)))))
    (with-alien ((alien-time time-t time))
      (let ((result (alien-funcall fn (addr alien-time))))
        (values (slot result 'tm-sec)              ; seconds
                (slot result 'tm-min)              ; minutes
                (slot result 'tm-hour)              ; hours
                (slot result 'tm-mday)              ; day of month
                (slot result 'tm-mon)              ; month
                (slot result 'tm-year)              ; year
                (slot result 'tm-wday)              ; day of week
                (slot result 'tm-yday)              ; day of year
                (slot result 'tm-isdst))))))            ; daylight savings time

(defun libc-localtime (time)
  (declare (type integer time))
  (let ((fn (extern-alien "localtime" (function (* (struct time)) (* time-t)))))
    (with-alien ((alien-time time-t time))
      (let ((result (alien-funcall fn (addr alien-time))))
        (values (slot result 'tm-sec)              ; seconds
                (slot result 'tm-min)              ; minutes
                (slot result 'tm-hour)              ; hours
                (slot result 'tm-mday)              ; day of month
                (slot result 'tm-mon)              ; month
                (slot result 'tm-year)              ; year
                (slot result 'tm-wday)              ; day of week
                (slot result 'tm-yday)              ; day of year
                (slot result 'tm-isdst))))))            ; daylight savings time

(defun gmt-offset (time)
  (let* ((gmt (multiple-value-list (libc-gmtime time)))
         (loc (multiple-value-list (libc-localtime time)))
         (delta (mapcar #'- gmt loc)))
    (* -3600 (if (not (= 0 (nth 7 delta)))
                (+ 24 (nth 2 delta))
                (nth 2 delta)))))

(defun utc-offset ()
  (gmt-offset (nth 1 (multiple-value-list (unix:unix-gettimeofday)))))

(def-alien-variable ("timezone" libc-timezone) time-t)

(def-alien-variable ("tzname" libc-tzname) (array c-string 2))

(defun libc-mktime (year mon mday &optional (hour 0) (min 0) (sec 0))
  (declare (type integer year mon mday hour min sec))
  (let ((fn (extern-alien "mktime" (function time-t (* (struct time))))))
    (with-alien ((alien-time (struct time)))
      (setf (slot alien-time 'tm-year) (- year 1900))
      (setf (slot alien-time 'tm-mon) (1- mon))
      (setf (slot alien-time 'tm-mday) mday)
      (setf (slot alien-time 'tm-hour) hour)
      (setf (slot alien-time 'tm-min) min)
      (setf (slot alien-time 'tm-sec) sec)
      (setf (slot alien-time 'tm-yday) 0)
      (setf (slot alien-time 'tm-wday) 0)
      (setf (slot alien-time 'tm-isdst) 0)
      (let ((result (alien-funcall fn (addr alien-time))))
        result))))
||#