;;; -*- Mode: Lisp -*-
;;; Ystok-Local-Time - Parsing ISO8601 and local formats
;;; Copyright (c) 2002-2014 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :ystok.local-time)


(defun valid-zone (sign hours minutes)
  (unless (<= 0 hours 12)
    (error 'iso-8601-parse-error :component-name :timezone-hour :bad-token hours))
  (when (and minutes (not (<= 0 minutes 59)))
    (error 'iso-8601-parse-error :component-name :timezone-minute :bad-token minutes))
  (* (if (eql sign #\-) -1 1)
     (+ (* hours 3600) (* (or minutes 0) 60))))

(defun valid-time-p (hours minutes seconds &optional msec suffix
                                                     (error 'local-time-parse-error))
 ;;; Args: suffix  If number, specifies the upper bound for an hour fragment.
  (cond ((not (and hours (<= 0 hours (cond ((numberp suffix) suffix)
                                           (suffix 12)
                                           (t 23)))))
         (error error :component-name #-Russian :hour #+Russian "''"
                      :bad-token hours))
        ((and minutes (not (<= 0 minutes (if (eql hours 24) 0 59))))
         (error error :component-name #-Russian :minute #+Russian "''"
                      :bad-token minutes))
        ((and seconds (not (<= 0 seconds (if (eql hours 24) 0 59))))
         (error error :component-name #-Russian :second #+Russian "''"
                      :bad-token seconds))
        ((and msec (not (<= 0 msec 999)))
         (error error :component-name #-Russian :millisecond #+Russian "''"
                      :bad-token msec))
        (t)))

(eval-when (:compile-toplevel :load-toplevel :execute) (ystok.meta:enable-meta-syntax))

(defun meta-parse-iso-8601 (string &key (start 0) end junk-allowed
                                        (date-elements t) (time-elements t)
                                        (timezone-p t))
 ;;; Args: date-elements,time-elements,timezone-p - if null,the component is not allowed
  ;; Values: index, year, month, day, hour, minute, second, msec, zone
  ;;	     When month < 0, it is equal to (- week).
 (flet ((decimal-separator (char) (member char '(#\, #\.))))
  (let (year month day
        hours minutes seconds msec
        zhours zminutes zsign
        date time zone 
        ahead)						; look ahead buffer
   (ystok.meta:with-string-meta (string :start start :end end)
     (labels
      ((%date (&aux (old-index ystok.meta:%index%) c)
         ;(nilf year month day)			; needed for rollback
         (or (ystok.meta:match
              [ @(digit c) !(setq year (ctoi c))	; year obligatory DI 2009-Oct-01?
                $[@(digit c) !(setq year (+ (* (or year 0) 10) (ctoi c)))]
                #\- !(push #\- ahead)
                {[#\W !(push #\W ahead) @(digit c) !(setq ahead nil month (- (ctoi c)))
                  [@(digit c) !(setq month (+ (* month 10) (- (ctoi c))))]]
                 [@(digit c) !(setq ahead nil month (ctoi c))
                  [@(digit c) !(setq month (+ (* month 10) (ctoi c)))]] }
                #\- !(push #\- ahead) 
                @(digit c) !(setq ahead nil day (ctoi c))
                $[@(digit c) !(setq day (+ (* day 10) (ctoi c)))] ])
             (setq ystok.meta:%index% old-index
                   ahead nil)))
       (%time (&aux (old-index ystok.meta:%index%) c (msec-denominator 1))
         ;(nilf hours minutes seconds msec)	; needed for rollback
         (if (ystok.meta:match
              [ @(digit c) !(setq hours (ctoi c))
                {[@(digit c) !(setq hours (+ (* hours 10) (ctoi c)))] []}
                {[#\: !(push #\: ahead) @(digit c) !(setq ahead nil minutes (ctoi c))
                  {[@(digit c) !(setq minutes (+ (* minutes 10) (ctoi c)))] []}
                  {[#\: !(push #\: ahead) @(digit c) !(setq ahead nil seconds (ctoi c))
                    {[@(digit c) !(setq seconds (+ (* seconds 10) (ctoi c)))] []}
                    {[@(decimal-separator c) !(push c ahead)
                      $[@(digit c) !(setq ahead nil
                                          msec (+ (* (or msec 0) 10) (ctoi c))
                                          msec-denominator (* msec-denominator 10))]]
                     []}]
                   []}]
                 []} ])
             (cond (ahead (setq ystok.meta:%index% old-index) nil)
                   (msec  (setq msec (truncate (* msec 1000) msec-denominator)))
                   (t))
             (progn (setq ystok.meta:%index% old-index) nil)))
       (%zone (&aux (old-index ystok.meta:%index%) c)
         (or (ystok.meta:match
              { [#\Z !(allf 0 zhours zminutes)]
                [@(sign zsign)
                 [@(digit c) !(setq zhours (ctoi c))]		; first hours digit
                 {[@(digit c) !(setq zhours (+ (* zhours 10) (ctoi c)))] []}
                 {[#\: !(push #\: ahead) @(digit c) !(setq ahead nil zminutes (ctoi c))
                   $[@(digit c) !(setq zminutes (+ (* zminutes 10) (ctoi c)))]]
                  []}] })
             (setq ystok.meta:%index% old-index
                   ahead nil))) )
       (cond ((cond ((and date-elements time-elements)	; both allowed
                     (if timezone-p
                         (ystok.meta:match
                          [ !(setq date (%date))
                            {[{#\T #\Space} !(setq time (%time)) {!(setq zone (%zone)) []}]
                            []} ])
                         (ystok.meta:match
                          [ !(setq date (%date))
                            {[{#\T #\Space} !(setq time (%time))] []} ])))
                    (date-elements			; only date allowed
                     (ystok.meta:match
                      [ !(setq date (%date)) ] ))
                    (time-elements			; only time allowed
                     (if timezone-p
                         (ystok.meta:match
                          [ !(setq time (%time)) {!(setq zone (%zone)) []} ])
                         (ystok.meta:match
                          [ !(setq time (%time)) ] ))) )
              (when date
                (if (minusp month)
                    (valid-ywd-date-p year (- month) day 'iso-8601-parse-error)
                    (valid-date-p year month day 'iso-8601-parse-error)))
                #|(unless (<= 1 year 9999)
                  (error 'iso-8601-syntax-error :component-name :year :bad-token year)|#
              (when time
                (valid-time-p hours minutes seconds msec nil 'iso-8601-parse-error))
              (when zone
                (setq zone (valid-zone zsign zhours zminutes)))
              (cond (junk-allowed)
                    (ahead 
                     (error 'iso-8601-parse-error :bad-token (coerce ahead 'string)))
                    ((< ystok.meta:%index% ystok.meta:%end%)
                     (error 'iso-8601-parse-error :bad-token
                            (subseq ystok.meta:%source% ystok.meta:%index%
                                    ystok.meta:%end%))))
              (values ystok.meta:%index%
                      year month day hours minutes seconds msec zone))
	     ;; Does not match
             (junk-allowed
              (values ystok.meta:%index% nil))
             (t (error 'iso-8601-parse-error :bad-token
                       (subseq ystok.meta:%source% start ystok.meta:%end%)))
)))) ) )

(defun meta-parse-local-time (string &key (start 0) end junk-allowed
                                     (date-format *date-format*)
                                     (date-separator *date-separator*)
                                     (century-digits *century-digits*)
                                     (time-format *time-format*)
                                     (time-separator *time-separator*)
                                     (decimal-separator *decimal-separator*))
 ;;; Args: junk-allowed Additionally permits empty year and/or month components,
  ;;			which could be defaulted to *default-year* and *default-month*
 (flet ((date-separator (char) (char= char date-separator))
        (time-separator (char) (char= char time-separator))
        (decimal-separator (char) (char= char decimal-separator)))
  (let (year month day 
        hours minutes seconds msec suffix
        zhours zminutes zsign
        date time zone 
        ahead)						; look ahead buffer
   (ystok.meta:with-string-meta (string :start start :end end)
     (labels 
      ((%date (&aux (old-index meta:%index%) c)
         ;(nilf year month day)				; needed for rollback
         (or (case date-format
               ((:dmy 1)
                (ystok.meta:match
                 [ @(digit c) !(setq day (ctoi c))		; day is obligatory
                   $[@(digit c) !(setq day (+ (* day 10) (ctoi c)))]
                   @(date-separator c) !(push c ahead)
                   $[@(digit c) !(setq ahead nil		; month is optional
                                       month (+ (* (or month 0) 10) (ctoi c)))]
                   @(date-separator c) !(push c ahead)
                   $[@(digit c) !(setq ahead nil		; year is optional
                                       year (+ (* (or year 0) 10) (ctoi c)))] ]))
               ((:ymd 2)
                (ystok.meta:match
                 [ $[@(digit c) !(setq year (+ (* (or year 0) 10) (ctoi c)))]
                   @(date-separator c) !(push c ahead)
                   $[@(digit c) !(setq ahead nil
                                       month (+ (* (or month 0) 10) (ctoi c)))]
                   @(date-separator c) !(push c ahead)
                   @(digit c) !(setq ahead nil day (ctoi c))	; day is obligatory
                   $[@(digit c) !(setq day (+ (* day 10) (ctoi c)))] ]))
               ((:mdy 0)
                (ystok.meta:match
                 [ $[@(digit c) !(setq month (+ (* (or month 0) 10) (ctoi c)))]
                   @(date-separator c) !(push c ahead)
                   @(digit c) !(setq ahead nil day (ctoi c))	; day is obligatory
                   $[@(digit c) !(setq day (+ (* day  10) (ctoi c)))]
                   @(date-separator c) !(push c ahead)
                   $[@(digit c) !(setq ahead nil
                                       year (+ (* (or year 0) 10) (ctoi c)))] ]))
               )
             (setq meta:%index% old-index
                   ahead nil)))
       (%time (&aux (old-index ystok.meta:%index%) c (msec-denominator 1))
         (nilf hours minutes seconds msec)
         (if (ystok.meta:match
              [ @(digit c) !(setq hours (ctoi c))
                {[@(digit c) !(setq hours (+ (* hours 10) (ctoi c)))] []}
                {[@(time-separator c) !(push c ahead)
                  @(digit c) !(setq ahead nil minutes (ctoi c))
                  {[@(digit c) !(setq minutes (+ (* minutes 10) (ctoi c)))] []}
                  {[@(time-separator c) !(push c ahead)
                    @(digit c) !(setq ahead nil seconds (ctoi c))
                    {[@(digit c) !(setq seconds (+ (* seconds 10) (ctoi c)))] []}
                    {[@(decimal-separator c) !(push c ahead)
                      $[@(digit c) !(setq ahead nil
                                          msec (+ (* (or msec 0) 10) (ctoi c))
                                          msec-denominator (* msec-denominator 10))]]
                     []}]
                   []}]
                 []} ])
             (cond (ahead (setq ystok.meta:%index% old-index) nil)
                   (msec  (setq msec (truncate (* msec 1000) msec-denominator)))
                   (t))
             (progn (setq ystok.meta:%index% old-index) nil)))
       (%zone (&aux (old-index ystok.meta:%index%) c)
         (or (ystok.meta:match
              { [#\Z !(allf 0 zhours zminutes)]
                [@(sign zsign)
                 [@(digit c) !(setq zhours (ctoi c))]		; first hours digit
                 {[@(digit c) !(setq zhours (+ (* zhours 10) (ctoi c)))] []}
                 {[#\: !(push #\: ahead) @(digit c) !(setq ahead nil zminutes (ctoi c))
                   $[@(digit c) !(setq zminutes (+ (* zminutes 10) (ctoi c)))]]
                  []}] })
             (setq ystok.meta:%index% old-index
                   ahead nil))) )
       (cond ((cond ((and date-format time-format)
                     (ystok.meta:match
                      [ !(setq date (%date))
                         {[{#\T #\Space} !(setq time (%time)) {!(setq zone (%zone)) []}
                           #1={[!(eql time-format 12) {#\Space []}
                               {["AM" !(setq suffix :am)]
                                ["PM" !(setq suffix :pm)]}]
                               []}
                          ] []}
                      ]))
                    (date-format
                     (ystok.meta:match
                      [ !(setq date (%date)) ]))
                    (time-format
                     (ystok.meta:match
                      [ !(setq time (%time)) {!(setq zone (%zone)) []}
                        #1#
                      ])))
              (when date
                ;; Interpret a la Oracle RR year with 2 digits:
                ;; - If the year is <50 and the last 2 digits of the current
                ;;   year are >=50, the first 2 digits of the returned year
                ;;   are 1 greater than the first two digits of the current year.
                ;; - If the year is >=50 and the last 2 digits of the
                ;;   current year are <50, the first 2 digits of the returned year
                ;;   are 1 less than the first two digits of the current year.
                ;; - Otherwise the first 2 digits of the returned year are the same
                ;;   as the first 2 digits of the current year.                
                (unless (or year (and junk-allowed (setq year *default-year*)))
                  (error 'local-time-parse-error :component-name :year :bad-token string))
                (unless (or month (and junk-allowed (setq month *default-month*)))
                  (error 'local-time-parse-error :component-name :month :bad-token string))
                (when (and (<= 0 year 99) (eql century-digits 2))
                  (multiple-value-bind (century yy) (floor (current-year) 100)
                    (incf year (cond ((and (< year 50) (>= yy 50))
                                      (* 100 (1+ century)))
                                     ((and (>= year 50) (< yy 50))
                                      (* 100 (1- century)))
                                     (t (* 100 century))) )))
                (valid-date-p year month day 'local-time-parse-error))
              (when time
                (valid-time-p hours minutes seconds msec suffix 'local-time-parse-error)
                (case suffix
                  (:am (when (eql hours 12) (setq hours 0)))
                  (:pm (when (< hours 12) (incf hours 12)))))
              (when zone
                (setq zone (valid-zone zsign zhours zminutes)))
              (cond (junk-allowed)
                    (ahead 
                     (error 'local-time-parse-error :bad-token (coerce ahead 'string)))
                    ((< ystok.meta:%index% ystok.meta:%end%)
                     (error 'local-time-parse-error :bad-token
                            (subseq ystok.meta:%source% ystok.meta:%index%
                                    ystok.meta:%end%))))
              (values ystok.meta:%index%
                      year month day hours minutes seconds msec zone))
	     ;; Does not match
             (junk-allowed
              (values ystok.meta:%index% nil))
             (t (error 'local-time-parse-error :bad-token
                       (subseq ystok.meta:%source% start ystok.meta:%end%)))
)))) ) )

(defun parse-duration (string &key (start 0) end junk-allowed (ym t) (dt t))
 ;;; Args: ym, dt - allow or disallow year-month or day-time durations
  ;; Values: duration,index
 (flet ((decimal-separator (char) (member char '(#\, #\.))))
  (let (sign				; + or -
        year month 
        week day hours minutes seconds
        floatp)				; a float component was matched last
   (ystok.meta:with-string-meta (string :start start :end end)
     (labels (
       (component (designator &aux (old-index ystok.meta:%index%) c d
                                   (f 0) (denominator 1))
         ;(prog1 
         (cond (floatp nil)		; no components allowed after a previous float
               ((ystok.meta:match
                 [ @(digit c) !(setq d (ctoi c))
                    $[@(digit c) !(setq d (+ (* d 10) (ctoi c)))]
                    {[@(decimal-separator c) 
                       $[@(digit c) !(setq f (+ (* f 10) (ctoi c))
                                           denominator (* denominator 10))]]
                     []}
                    @(alpha-char-p c) !(char-equal c designator) ])
                (if (eql f 0)
                    d
                    (progn (setq floatp t)
                      (+ d (/ (float f) denominator)))))
               (t (setq ystok.meta:%index% old-index) nil)))
          ;   (format t "~&component:d ~s f ~s denominator ~s" d f denominator)))
       )
       (cond ((ystok.meta:match
               [ {@(sign sign) []}
                 {[!ym !(setq year (component #\Y))
                       {!(setq month (component #\M)) []} ]
                  [!ym !(setq month (component #\M))
                       !(if (component #\S)    ; rollback entirely in favour of dt
                            (setq ystok.meta:%index% start  month nil  floatp nil)
                            t) ]
                  [!dt !(setq week (component #\W))
                       {!(setq day (component #\D)) []}
                       {#\T []}
                       {!(setq hours (component #\H)) []}
                       {!(setq minutes (component #\M)) []}
                       {!(setq seconds (component #\S)) []} ] 
                  [!dt !(setq day (component #\D))
                       {#\T []}
                       {!(setq hours (component #\H)) []}
                       {!(setq minutes (component #\M)) []}
                       {!(setq seconds (component #\S)) []} ] 
                  [!dt #\T
                       {!(setq hours (component #\H)) []}
                       {!(setq minutes (component #\M)) []}
                       {!(setq seconds (component #\S)) []} ] 
                  [!dt !(setq hours (component #\H))
                       {!(setq minutes (component #\M)) []}
                       {!(setq seconds (component #\S)) []} ] 
                  [!dt {!(setq minutes (component #\M)) []}
                       !(setq seconds (component #\S)) ] 
                 } ])
              (let* ((minusp (eql sign #\-))
                     (duration
                      (cond ((and (or year month) (not (or day hours minutes seconds)))
                             (encode-ym-duration
                              (if year (if minusp (- year) year) 0)
                              (if month (if minusp (- month) month) 0)))
                            ((and (not (or year month))
                                  (or week day hours minutes seconds))
                             ;; Signal if more the last float component supplied.
                             ;; Junk-allowed does not work in this case.
                             ;(let ((list (list week day hours minutes seconds)))
                             ;  (when-bind (pos-float (position-if #'floatp list))
                             ;    (let ((pos (position-if-not #'null list :from-end t)))
                             ;      (when (< pos-float pos)
                             ;        (error 'duration-parse-error
                             ;               :component-name (nth pos-float
                             ;                      '(:week :day :hour :minute :second))
                             ;               :bad-token (nth pos-float list))))))
                             (encode-dt-duration
                              (+ (if week (* (if minusp (- week) week) 7) 0)
                                 (if day (if minusp (- day) day) 0))
                              (if hours (if minusp (- hours) hours) 0)
                              (if minutes (if minusp (- minutes) minutes) 0)
                              (if seconds (if minusp (- seconds) seconds) 0)
                              ))	; check:=nil
                            (t (error 'duration-parse-error :bad-token
                                      (subseq ystok.meta:%source% start
                                              ystok.meta:%end%)) ))))
                (when (and (not junk-allowed)
                           (< ystok.meta:%index% ystok.meta:%end%))
                  (error 'duration-parse-error :bad-token
                         (subseq ystok.meta:%source% ystok.meta:%index%
                                 ystok.meta:%end%)))
                (values duration ystok.meta:%index%)))

	     ;; Does not match
             (junk-allowed
              (values nil ystok.meta:%index%))
             (t (error 'duration-parse-error :bad-token
                  (subseq ystok.meta:%source% ystok.meta:%index% ystok.meta:%end%))))
         
    )))))

(eval-when (:compile-toplevel :load-toplevel :execute) (ystok.meta:disable-meta-syntax))

(setf (fdefinition 'meta-parse-duration) (fdefinition 'parse-duration))	; for compatibility

(defun parse-timestring (timestring &rest args
                                    &key local (start 0) end junk-allowed
                                         (date-format *date-format*)
                                         (time-format *time-format*)
                                         ym dt 
                                    &allow-other-keys)
  "Parse a timestring and return the corresponding date or local-time object.
Arguments:
 TIMESTRING  If starts with P, read a ISO 8601 duration.
 YM or DT    If specified, read a localized duration,
 LOCAL       If true, read a localized formatted date/time string;
             otherwise read an ISO 8601 formatted date/time string.
 TIME-FORMAT When eql to :auto, return a date instance if none of
             the time components is found in the timestring."
  ;; Implementation Note: only dt-duration up to 24 hours can be read localized for now.
  (cond ((char= (char timestring start) #\P)
         (parse-duration timestring :start (1+ start) :end end :junk-allowed junk-allowed))
        ((or ym dt)
         (let ((minusp (char= (char timestring start) #\-)))
           (multiple-value-bind (position year month day hours minutes seconds msec zone)
             (apply #'meta-parse-local-time timestring
                    :start (if minusp (1+ start) start)
                    :date-format nil  :time-format 24
                    (remove-properties args '(:local :ym :dt)))
             (declare (ignore year month zone))
             (values 
              (encode-dt-duration (if day (if minusp (- day) day) 0)
                                  (if hours (if minusp (- hours) hours) 0)
                                  (if minutes (if minusp (- minutes) minutes) 0)
                                  (if seconds (if minusp (- seconds) seconds) 0)
                                  (if msec (if minusp (- msec) msec) 0)
                                  nil)		; check:=nil
              position))))
        ((multiple-value-bind (position year month day hours minutes seconds msec zone)
          (if local
              (apply #'meta-parse-local-time timestring
                     (remove-properties args '(:local)))
              (meta-parse-iso-8601 timestring :start start :end end
                                   :junk-allowed junk-allowed
                                   :date-elements date-format
                                   :time-elements time-format
                                   ))
           (values
            ;(when year		; there is no pure time concept (without day component)
            (if (and time-format
                     (imp (eq time-format :auto) (or msec seconds minutes hours)))
                (encode-local-time (or msec 0)
                                   (or seconds 0) (or minutes 0) (or hours 0)
                                   (or day *default-day*)
                                   (or month *default-month*)
                                   (or year *default-year*)
                                   (or zone *session-timezone*)
                                   (if zone 'utc-time *default-encode-type*)
                                   nil)	; don't check
                (encode-date (or year *default-year*)
                             (or month *default-month*)
                             (or day *default-day*)
                             nil))		; don't check
            position)))))

(defun iso-8601-time-reader (stream character)
  (parse-timestring
   (if (eql (setq character (peek-char nil stream nil nil t)) #\")
       (read stream t nil t)			; must be a string
       (coerce (loop for char = character then (peek-char nil stream nil nil t)
                     while (and char (or (alphanumericp char) (find char "-:,.+")))
                     collect (read-char stream t nil t))
               'string))))

(set-macro-character #\@ 'iso-8601-time-reader t)
;(set-macro-character #\@ nil t) ; disable

(defun universal-time-reader (stream subchar arg)
  (declare (ignore arg))
  (let ((local-time (iso-8601-time-reader stream subchar)))
    (universal-time local-time)))

(set-dispatch-macro-character #\# #\@ 'universal-time-reader)
;(set-dispatch-macro-character #\# #\@ nil) ; disable

