;;; -*- Mode: Lisp -*-
;;; Ystok-Local-Time - Pringing ISO8601 and local formats
;;; Copyright (c) 2002-2014 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(in-package :ystok.local-time)

;;; ISO 8601

(defun format-timestring (stream arg &optional
                          universal-p                    ; colon
                          timezone-p                     ; atsign
                          date-elements time-elements
                          date-separator time-separator 
                          internal-separator)
 "Produces on stream the timestring corresponding to the time with the options given.
 Owing to |~@| synonym, can be used in (format nil \"~/lt:~@/\" arg)"
 (flet ((format-iso-8601 (stream)
   (let (msec seconds minutes hours day month year dow dst displacement)
    (cond ((local-time-p arg)
           (if universal-p
               (multiple-value-setq (msec seconds minutes hours day month year)
                   (decode-as-utc arg))
                                          ; output according to *session-timezone*
               (multiple-value-setq (msec seconds minutes hours day month year dow
                                     displacement dst)
                   (decode-as-local arg timezone-p)))
                 ;(when zone (setq  (+ zone (or dst 0))))))
           (setq msec (local-time-msec arg)))
          ((date-p arg)
           (multiple-value-setq (year month day) (decode-date arg))
           (setq time-elements 0))
          ;; Universal time
          ((integerp arg)
           (multiple-value-setq (seconds minutes hours day month year dow
                                 dst displacement)
               (cond (universal-p
                      (decode-universal-time arg 0))
                     ((member *session-timezone* '(t nil))
                      (decode-universal-time arg))
                     ((numberp *session-timezone*)
                      (decode-universal-time arg (/ *session-timezone* -3600)))))
           ;; As displacement was returned as standard in hours, convert to reciprocal
           (setq msec 0
                 dow (1+ dow)		; as dow was returned: 0=mon,...,6=sun
                 displacement (* (- (if dst 1 0) displacement) 3600)))
          (t
           (error 'type-error :datum arg :expected-type '(or local-time integer))))
            
    (unless date-elements (setq date-elements 4))
    (unless time-elements (setq time-elements 4))    ; conforming: default should be 4
    (cond ((eql date-separator 0) (setq date-separator nil)) ; basic format
          ((null date-separator) (setq date-separator #\-)))
    (cond ((eql date-elements 0))
          ((keywordp date-elements)
           (cond ((eq date-elements :yw) (setq dow nil))
                 ((null dow) (setq dow (time-component arg :week-day-number universal-p))))
           (multiple-value-bind (week year) (week arg t)
             (format stream "~4,'0D~@[~C~]W~2,'0D~:[~;~@[~C~]~D~]"
                     year date-separator week dow date-separator dow)))
          (t
           (case date-elements
             (4 (format stream "~4,'0D" year))
             (3 (format stream "~2,'0D" (mod year 100))) ;omit century 
             (otherwise (write-char #\- stream)))
           (format stream "~@[~C~]~:[~2*~;~2,'0D~@[~C~]~]~2,'0D"
                   date-separator (>= date-elements 2) month
                   date-separator day)))

    (unless (eql time-elements 0)                     ; internal separator needed?
        (cond ((eql internal-separator 0))
              ((and (eql date-elements 0) (null internal-separator)))
              (t (write-char (or internal-separator *internal-separator*) stream)))
        (cond ((eql time-separator 0) (setq time-separator nil)) ; basic format
              ((null time-separator) (setq time-separator #\:)))
        (when (>= time-elements 1)
          (format stream "~2,'0D" hours))
        (when (>= time-elements 2)
          (format stream "~@[~C~]~2,'0D" time-separator minutes))
        (when (>= time-elements 3)
          (format stream "~@[~C~]~2,'0D" time-separator seconds))
        (when (and (/= msec 0) (>= time-elements 4))
          (format stream "~C~3,'0D" *decimal-separator* msec)))
    (when timezone-p 
      (cond (universal-p
             (write-char #\Z stream))
            (displacement
             (multiple-value-bind (tz-hour tz-sec) (floor (abs displacement) 3600)
               (let ((tz-min (floor tz-sec 60)))
                 (format stream "~:[+~;-~]~2,'0D~:[:~2,'0D~;~]"
                         (minusp displacement) tz-hour
                         (zerop tz-min) tz-min)))))))
   ))
   (if stream
       (format-iso-8601 stream)
       (with-output-to-string (stream)
         (format-iso-8601 stream)))))

(defmethod print-object ((self date) stream)
  ;(if (*print-readably*))
  (when *print-escape* (write-char #\@ stream))
  (format-timestring stream self nil t))

;;; Enable using as a format specifier. Pattern: 
;;;	(format nil "~,3,,,' @/lt:~@/" (get-local-time))
(export '|~@| :local-time)
(setf (fdefinition '|~@|) (fdefinition 'format-timestring))

(defmacro \@-format-to-string (arg &key 
                                   universal                    ; colon
                                   timezone                     ; atsign
                                   date-elements time-elements
                                   date-separator time-separator 
                                   internal-separator)
  `(funcall #'format-timestring nil ,arg ,universal ,timezone
                                         ,date-elements ,time-elements
                                         ,date-separator ,time-separator 
                                         ,internal-separator))

;;; LOCALIZED

(defun ~ts (stream arg &optional universal-p                    ; colon
                                 timezone-p                     ; atsign
                                 date-format century-digits date-separator
                                 time-format time-elements time-separator)
  ;; date-format ::= (0 = :mdy) (1 = :dmy) (2 = :ymd) :none;
  ;;		otherwise (:ywd or :yw) passed to format-timestring
  ;; time-format ::= 12 or 24
  (unless date-format (setq date-format *date-format*))
  (unless century-digits (setq century-digits *century-digits*))
  (unless date-separator (setq date-separator *date-separator*))
  (unless time-format (setq time-format *time-format*))
  (unless time-elements (setq time-elements 2))
  (unless time-separator (setq time-separator *time-separator*))
  (let (msec seconds minutes hours day month year dow displacement dst)
    (declare (ignorable dow))
    (cond ((local-time-p arg)
           (multiple-value-setq (msec seconds minutes hours day month year dow
                                 displacement)
               (if universal-p
                   (decode-as-utc arg)
                   (decode-as-local arg timezone-p)))
           (setq msec (local-time-msec arg)))
          ((integerp arg)					; Lisp universal time
           (multiple-value-setq (seconds minutes hours day month year dow
                                 dst displacement)
               (cond (universal-p
                      (decode-universal-time arg 0))
                     ((member *session-timezone* '(t nil))
                      (decode-universal-time arg))
                     ((numberp *session-timezone*)
                      (decode-universal-time arg (/ *session-timezone* -3600)))))
           ;; As displacement was returned as standard in hours, convert to reciprocal
           (setq msec 0
                 displacement (* (- (if dst 1 0) displacement) 3600)))
          ((date-p arg)
           (setq time-elements 0)
           (multiple-value-setq (year month day) (%lt-day-ymd (local-time-day arg))))
          (t
           (error 'type-error :datum arg :expected-type '(or local-time integer))))

    (case century-digits
      (2 (setq year (nth-value 1 (truncate year 100))))
      (3 (setq year (nth-value 1 (truncate year 1000))))
      (0 (setq year nil)))
    (case date-format
      ((:dmy 1)
       (format stream "~2,'0D~C~2,'0D~:[~3*~;~C~v,'0D~]"
               day date-separator month year date-separator century-digits year))
      ((:ymd 2)
       (format stream "~:[~3*~;~v,'0D~C~]~2,'0D~C~2,'0D"
               year century-digits year date-separator month date-separator day))
      ((:mdy 0)
       (format stream "~2,'0D~C~2,'0D~:[~3*~;~C~v,'0D~]"
               month date-separator day year date-separator century-digits year))
      (:none)
      (:d-mon-y
       (format stream "~2,'0D~C~A~:[~3*~;~C~v,'0D~]"
               day date-separator (svref yl::*month-abbrevs* (1- month))
               year date-separator century-digits year))
      (otherwise
       (format-timestring stream arg universal-p timezone-p date-format time-elements
                          (if (eql date-separator 0) date-separator))
       (return-from ~ts)))

    (when (> time-elements 0)
      (unless (eq date-format :none)
        (write-char #\Space stream))
      (let (suffix)
        (cond ((eql time-format 24) (format stream "~D" hours)) ;"~2,'0D" DI 2008-Dec-22
              ((eql hours 0)  (write-string "12" stream) (setq suffix " AM"))
              ((eql hours 12) (write-string "12" stream) (setq suffix " PM"))
              ((< hours 12)   (format stream "~D" hours) (setq suffix " AM"))
              (t (format stream "~D" (- hours 12)) (setq suffix " PM"))) ;(< 12 hour 24)

        (when (>= time-elements 2)
          (format stream "~C~2,'0D" time-separator minutes)
          (when (>= time-elements 3)
            (format stream "~C~2,'0D" time-separator seconds)
            (when (and (/= msec 0) (>= time-elements 4))
              (format stream "~C~3,'0D" *decimal-separator* msec))))

        (when timezone-p 
          (cond (universal-p
                 (write-char #\Z stream))
                (displacement
                 (multiple-value-bind (tz-hour tz-sec) (floor (abs displacement) 3600)
                   (let ((tz-min (floor tz-sec 60)))
                     (format stream "~:[+~;-~]~2,'0D~:[:~2,'0D~;~]"
                             (minusp displacement) tz-hour
                             (zerop tz-min) tz-min))))))
        (when suffix
          (write-string suffix stream))))))

(defun ts-format-to-string (arg &key universal timezone
                                     date-format century-digits date-separator
                                     time-format time-elements time-separator)
  (if arg
      (with-output-to-string (stream)
        (~ts stream arg universal timezone
             date-format century-digits date-separator
             time-format time-elements time-separator))
      +null-string+))

(defun ~d (stream arg &optional universal-p atsign
                                date-format century-digits date-separator)
  (declare (ignore atsign))
  (~ts stream arg universal-p nil date-format century-digits date-separator 24 0))

(defun d-format-to-string (arg &key universal format century-digits separator)
  (if arg
      ;(with-gensyms (stream)
      (with-output-to-string (stream)
       (~d stream arg universal nil format century-digits separator))
      +null-string+))

(defun ~t (stream arg &optional universal-p timezone-p
                                time-format time-elements time-separator)
  (~ts stream arg universal-p timezone-p
       :none 0 #\- time-format time-elements time-separator))

(defun t-format-to-string (arg &key universal timezone format elements separator)
  (if arg
      ;(with-gensyms (stream)
      (with-output-to-string (stream)
        (~t stream arg universal timezone format elements separator))
      +null-string+))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;  DURATION  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; ISO 8601

(defun format-duration (stream arg
                        &optional alternative-p		; colon - print a la local-time
                                  weeks-p)		; atsign
 (declare (ignore alternative-p weeks-p))
 (flet ((format-iso-8601 (stream)
  (if (ym-duration-p arg)
      (multiple-value-bind (year month) (decode-duration arg)
        (let ((hide-year (and (= year 0) (/= month 0))))
          (format stream "P~:[~DY~;~*~]~:[~DM~;~*~]"
                  hide-year year
                  (= month 0) (if hide-year month (abs month)))))
      ;; else dt-duration or a number
      (multiple-value-bind (day hours minutes seconds msec) (decode-duration arg)
        (declare (fixnum day hours minutes seconds msec))
        (cond ((= 0 day hours minutes seconds msec)
               (write-string "P0D" stream))
              (t
               (when (/= msec 0)					; DI 2014-Jan-22
                 (incf seconds (the fixnum (round msec 1000))))		; (* msec 0.001d0)
               (format stream
                     "P~:[~;-~]~:[~DD~;~*~]~:[~;T~]~:[~DH~;~*~]~:[~DM~;~*~]~:[~DS~;~*~]"
                     (or (< day 0) (< hours 0) (< minutes 0) (< seconds 0))
                     (= day 0) (abs day)
                     (= day hours seconds 0)
                     (= hours 0) (abs hours)
                     (= minutes 0) (abs minutes)
                     (= seconds 0) (abs seconds))))))))
   (if stream
       (format-iso-8601 stream)
       (with-output-to-string (stream)
         (format-iso-8601 stream)))))

(defmethod print-object ((self duration) stream)
  (when *print-escape* (write-char #\@ stream))
  (format-duration stream self))

;;; LOCALIZED

(defun ~dt (stream arg &optional colon atsign time-elements time-separator d)
 ;;; Args: colon  If true, provide ~f output of the real number of units determined
  ;;		  by time-elements: 0 days, 1 hours, 2 minutes, 3 seconds.
  (unless time-elements (setq time-elements 2))
  (unless time-separator (setq time-separator *time-separator*))
  (multiple-value-bind (day hours minutes seconds msec) (decode-duration arg)
    (declare (fixnum day hours minutes seconds msec))
    ;(with-dt (day sec msec) arg
    (cond ((= 0 day hours minutes seconds msec)
           (write-char #\0 stream))
          (colon
           (~f stream				; convert to the number of days
               (case time-elements
                 (0 (if (duration-p arg)
                        (+ day
                           (/ (duration-sec arg) +secs-per-day+)
                           (/ (duration-msec arg) +msecs-per-day+))
                        arg))
                 (1 (if (duration-p arg)	; convert to the number of hours
                        (+ (* day 24) hours
                           (/ minutes 60) (/ seconds 3600) (/ msec 3600000))
                        (* arg 24)))
                 (2 (if (duration-p arg)	; convert to the number of minutes
                        (+ (* day #.(* 24 60)) (* hours 60) minutes
                           (/ seconds 60) (/ msec 60000))
                        (* arg #.(* 24 60))))
                 (3 (if (duration-p arg)	; convert to the number of seconds
                        (+ (* day +secs-per-day+)
                           (duration-sec arg)
                           (* (duration-msec arg) 0.001))
                        (* arg +secs-per-day+))) )
               colon atsign nil d nil nil nil nil))
          (t
           (when (or (< day 0) (< hours 0) (< minutes 0) (< seconds 0) (< msec 0))
             (write-char #\- stream))
           (when (/= day 0) (format stream "~DD" (abs day)))
           (when (>= time-elements 2)
             (format stream "~D" (abs hours))
             (when (>= time-elements 2)
               (format stream "~C~2,'0D" time-separator (abs minutes))
               (when (>= time-elements 3)
                 (format stream "~C~2,'0D" time-separator (abs seconds))
                 (when (and (/= msec 0) (>= time-elements 4))
                   (format stream "~C~v,'0D" *decimal-separator* (or d 3) (abs msec)))))))
) ) ) 

(defun dt-format-to-string (arg &key float elements separator d atsign)
  (if arg
      (with-output-to-string (stream)
        (~dt stream arg float atsign elements separator d))
      +null-string+))

#-Russian
(defun d-long-princ-to-string (date &optional week-day (year t) (day t))
 ;;; Output month name in wording, e.g. "4 August, monday"
  ;; Args: week-day If true, print week-day name after comma.
  ;;       year	    If true, print year.
  ;;	   day      If false, print only month and year.
  (flet ((%concat (y month d &optional week-day-number)
           (format nil "~@[~d ~]~a~@[ ~4,'0d~]~@[, ~a~]"
                   (and day d)
                   (month-name month)
                   (and year y)
                   (and week-day-number (svref *weekday-names* (1- week-day-number))))))
    (if (and day week-day)
        (if (local-time-p date)
            (multiple-value-bind (msec seconds minutes hours day month year
                                  week-day-number)
                (decode-as-local date)
              (declare (ignore msec seconds minutes hours))
              (%concat year month day week-day-number))
            (multiple-value-bind (year month d)
                (decode-date date)
              (%concat year month d (week-day-number date))))
        (multiple-value-call #'%concat (decode-date date)))))

