;;; -*- Mode: Lisp -*-
;;; Ystok-Local-Time - Calls to Win32 API
;;; Copyright (c) 2002-2014 Dr. Dmitry Ivanov. All rights reserved.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Russia NB: For Moscow, Caucasus Standard Time can be set without auto adjusting clock.

(in-package :ystok.local-time)

(eval-when (:compile-toplevel :load-toplevel :execute)

(ffc:define-foreign-struct |SYSTEMTIME| 
  (wYear      ffc:WORD)		; year
  (wMonth     ffc:WORD)		; month; January = 1, February = 2, and so on
  (wDayOfWeek ffc:WORD)		; day of the week; Sunday = 0, Monday = 1, and so on
  (wDay       ffc:WORD)		; day of the month
  (wHour      ffc:WORD)		; hour
  (wMinute    ffc:WORD)		; minute
  (wSecond    ffc:WORD)		; second
  (wMilliseconds ffc:WORD))	; millisecond

(ffc:define-foreign-struct |FILETIME|
  (low  ffc:DWORD)		; DWORD dwLowDateTime
  (high ffc:DWORD))		; DWORD dwHighDateTime

(ffc:define-foreign-struct |TIME_ZONE_INFORMATION| 
  (Bias :signed-long)	; current bias, minutes: UTC = local time + bias (Moscow -180)
  (StandardName (:c-array :wchar-t 32))	; Unicode string, e.g. "Russian Standard Time"
  (StandardDate |SYSTEMTIME|)		; wMonth=10,wDayOfWeek=0,wDay=5 (last Sun),wHour=3
  (StandardBias :signed-long)		; Usually 0
  (DaylightName (:c-array :wchar-t 32))	; Unicode string, e.g. "Russian Daylight Time"
  (DaylightDate |SYSTEMTIME|)		; wMonth=3,wDayOfWeek=0,wDay=5 (last Sun),wHour=2
  (DaylightBias :signed-long))		; bias value to be used during dst (usually -60)

(defconstant TIME_ZONE_ID_INVALID #xFFFFFFFF)
(defconstant TIME_ZONE_ID_UNKNOWN 0)	; no transition dates specified
(defconstant TIME_ZONE_ID_STANDARD 1)
(defconstant TIME_ZONE_ID_DAYLIGHT 2)
) ; eval-when

(ffc:define-win32api |GetSystemTime| ((ptr (:pointer |SYSTEMTIME|))) :void)
(ffc:define-win32api |GetLocalTime|  ((ptr (:pointer |SYSTEMTIME|))) :void)
(ffc:define-win32api |GetTimeZoneInformation| ((ptr (:pointer |TIME_ZONE_INFORMATION|)))
  ffc:DWORD)

(ffc:define-win32api |SystemTimeToTzSpecificLocalTime|
                   ; if lpTimeZone is NULL, uses the currently active time zone
                   ; (:reference-pass |TIME_ZONE_INFORMATION| :allow-null t)
  ((lpTimeZone (:pointer |TIME_ZONE_INFORMATION|))
   (lpSystemTime (:pointer |SYSTEMTIME|))
   (lpLocalTime  (:pointer |SYSTEMTIME|)))
  :boolean)

(ffc:define-win32api |FileTimeToSystemTime|
  ((lpFileTime   (:pointer |FILETIME|))		; IN CONST FILETIME *lpFileTime
   (lpSystemTime (:pointer |SYSTEMTIME|)))	; OUT LPSYSTEMTIME lpSystemTime
  :boolean)

(defmacro %decode-win32-system-time (ptr &optional (result-type 'values))
 ;;; Args: result-type ::= values | list
  (rebinding (ptr)
    `(,result-type (ffc:%slot ,ptr |SYSTEMTIME| wYear)
                   (ffc:%slot ,ptr |SYSTEMTIME| wMonth)	
                   (ffc:%slot ,ptr |SYSTEMTIME| wDay)
                   (ffc:%slot ,ptr |SYSTEMTIME| wHour)	
                   (ffc:%slot ,ptr |SYSTEMTIME| wMinute)
                   (ffc:%slot ,ptr |SYSTEMTIME| wSecond)
                   (ffc:%slot ,ptr |SYSTEMTIME| wMilliseconds)
                   (ffc:%slot ,ptr |SYSTEMTIME| wDayOfWeek))))

(defun get-win32-system-time ()
  "Return current UTC time"
  (ffc:with-foreign ((ptr |SYSTEMTIME|))
    (|GetSystemTime| ptr)
    (%decode-win32-system-time ptr)))

(defun get-win32-local-time ()
  "Return current UTC time"
  (ffc:with-foreign ((ptr |SYSTEMTIME|))
    (|GetLocalTime| ptr)
    (%decode-win32-system-time ptr)))

(defun get-win32-timezone-information ()
  (ffc:with-foreign ((ptr |TIME_ZONE_INFORMATION|))
    (let ((result (|GetTimeZoneInformation| ptr)))
      (case result
        (#.TIME_ZONE_ID_UNKNOWN
         (values result
                 (ffc:%slot ptr |TIME_ZONE_INFORMATION| Bias)
                 (ffc:%slot ptr |TIME_ZONE_INFORMATION| StandardBias)
                 (ffc:%slot ptr |TIME_ZONE_INFORMATION| DaylightBias)
                 nil nil		; no transition dates used for this timezone
                 (ffc:%cstring-lisp (ffc:%slot-pointer ptr |TIME_ZONE_INFORMATION|
                                                       StandardName))
                 nil))		    ;:external-format :unicode
        ((#.TIME_ZONE_ID_STANDARD #.TIME_ZONE_ID_DAYLIGHT)
         (values result
                 (ffc:%slot ptr |TIME_ZONE_INFORMATION| Bias)
                 (ffc:%slot ptr |TIME_ZONE_INFORMATION| StandardBias)
                 (ffc:%slot ptr |TIME_ZONE_INFORMATION| DaylightBias)
                 (%decode-win32-system-time
                  (ffc:%slot-pointer ptr |TIME_ZONE_INFORMATION| StandardDate) list)
                  ;(multiple-value-list (%decode-win32-system-time
                                       ;(fli:foreign-slot-value tzi 'StandardDate
                                       ;      :copy-foreign-object nil)))
                 (%decode-win32-system-time
                  (ffc:%slot-pointer ptr |TIME_ZONE_INFORMATION| DaylightDate) list)
                  ;(multiple-value-list (%decode-win32-system-time
                                       ;(fli:foreign-slot-value tzi 'DaylightDate
                                       ;      :copy-foreign-object nil)))
                 (ffc:%cstring-lisp (ffc:%slot-pointer ptr |TIME_ZONE_INFORMATION|
                                                       StandardName))
                 (ffc:%cstring-lisp (ffc:%slot-pointer ptr |TIME_ZONE_INFORMATION|
                                                       DaylightName))
                 ;(fli:convert-from-foreign-string
                 ; (fli:foreign-slot-pointer tzi 'StandardName) :external-format :unicode)
                 ;(fli:convert-from-foreign-string
                 ; (fli:foreign-slot-pointer tzi 'DaylightName) :external-format :unicode)
                 ))
        (otherwise (values))))))

(defun win32-timezone ()
 ;;; Values: 1) Current timezone displacement in seconds
  ;;         2) dst-p
  (ffc:with-foreign ((ptr |TIME_ZONE_INFORMATION|))
    (case (|GetTimeZoneInformation| ptr)
      ((#.TIME_ZONE_ID_UNKNOWN #.TIME_ZONE_ID_STANDARD)
       (values (* (+ (ffc:%slot ptr |TIME_ZONE_INFORMATION| Bias)
                     (ffc:%slot ptr |TIME_ZONE_INFORMATION| StandardBias))
                  -60)
               nil))
      (#.TIME_ZONE_ID_DAYLIGHT
       (values (* (+ (ffc:%slot ptr |TIME_ZONE_INFORMATION| Bias)
                     (ffc:%slot ptr |TIME_ZONE_INFORMATION| DaylightBias))
                  -60)
               t)))))

